#    _   _ _ _____ _    _              _____     _ _     ___ ___  _  __
#   /_\ | | |_   _| |_ (_)_ _  __ _ __|_   _|_ _| | |__ / __|   \| |/ /
#  / _ \| | | | | | ' \| | ' \/ _` (_-< | |/ _` | | / / \__ \ |) | ' <
# /_/ \_\_|_| |_| |_||_|_|_||_\__, /__/ |_|\__,_|_|_\_\ |___/___/|_|\_\
#                             |___/
#
# Copyright 2017 AllThingsTalk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AssetMismatchException(Exception):
    '''Existing asset on the Platform does not match the asset with the same name supplied in a device definition.'''
    pass



class Profile(dict):
    _type = None
    extras = []

    def __init__(self, **kwargs):
        self.type = self.__class__._type
        profile = {'type': self.type}
        for key, val in kwargs.items():
            if key in self.__class__.extras:
                profile[key] = val
        dict.__init__(self, **profile)


class NumberProfile(Profile):
    _type = 'number'
    extras = ['minimum', 'maximum', 'unit']


class IntegerProfile(Profile):
    _type = 'integer'
    extras = ['minimum', 'maximum', 'unit']


class BooleanProfile(Profile):
    _type = 'boolean'
    extras = []


class StringProfile(Profile):
    _type = 'string'
    extras = ['enum']


class ObjectProfile(Profile):
    _type = 'object'
    extras = ['properties']


class GeoProfile(ObjectProfile):

    def __init__(self, **kwargs):
        super().__init__(properties={
            'latitude': NumberProfile(),
            'longitude': NumberProfile()
        })



#    _   _ _ _____ _    _              _____     _ _     ___ ___  _  __
#   /_\ | | |_   _| |_ (_)_ _  __ _ __|_   _|_ _| | |__ / __|   \| |/ /
#  / _ \| | | | | | ' \| | ' \/ _` (_-< | |/ _` | | / / \__ \ |) | ' <
# /_/ \_\_|_| |_| |_||_|_|_||_\__, /__/ |_|\__,_|_|_\_\ |___/___/|_|\_\
#                             |___/
#
# Copyright 2017 AllThingsTalk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['Device']

import datetime
from dateutil.parser import parse as parse_date

import json
import copy

from . import assets


# inspired by https://github.com/django/django/blob/master/django/db/models/base.py
class DeviceBase(type):
    def __new__(cls, name, bases, attrs):
        super_new = super().__new__

        # Only perform for Device subclases (not Device class itself)
        parents = [b for b in bases if isinstance(b, DeviceBase)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        new_attrs = {'__module__': attrs.pop('__module__')}

        # Python 3.6 support:  http://stackoverflow.com/questions/41343263
        classcell = attrs.pop('__classcell__', None)
        if classcell is not None:
            new_attrs['__classcell__'] = classcell
        new_class = super_new(cls, name, bases, new_attrs)

        #
        # Class enhancements
        #

        # Message handlers for state / feed / command / event

        new_class._handlers = {}
        new_class.state = DeviceBase.HandlerDecoratorCollection(new_class, 'state')
        new_class.feed = DeviceBase.HandlerDecoratorCollection(new_class, 'feed')
        new_class.command = DeviceBase.HandlerDecoratorCollection(new_class, 'command')
        new_class.event = DeviceBase.HandlerDecoratorCollection(new_class, 'event')

        # Asset transformations
        for name, asset in attrs.items():
            if isinstance(asset, assets.Asset):
                # Configure asset name from variable name
                if not asset.name:
                    asset.name = name
                if not asset.title:
                    asset.title = name.capitalize()
                asset._internal_id = name

                # Create the actuation decorator
                new_class.state._add_asset(asset)
                new_class.feed._add_asset(asset)
                new_class.command._add_asset(asset)
                new_class.event._add_asset(asset)

        new_class._assets = [value for name, value in attrs.items()
                             if isinstance(value, assets.Asset)]

        return new_class

    class HandlerDecoratorCollection:

        def __init__(self, device_class, stream):
            device_class._handlers[stream] = {}
            self._stream = stream
            self._device_class = device_class
            self._assets = {}

        def _add_asset(self, asset):
            def decorator(fn):
                self._device_class._handlers[self._stream][asset.name] = fn
                return fn
            self._assets[asset._internal_id] = decorator

        def __getattr__(self, internal_id):
            if internal_id in self._assets:
                return self._assets[internal_id]
            else:
                return AttributeError


class Device(metaclass=DeviceBase):
    '''Device contains information about assets. It maps to AllThingsTalk
    Platform device resources.'''

    def __init__(self, *, client=None, id=None, connect=True,
                 overwrite_assets=False, **kwargs):
        '''Initializes the device

        :param Client client: The client used to interface with the platform
        :param str id: Device resource id. If supplied, the device will be mapped to the device resource. If None, an attempt will be made to create the device.
        :param connect boolean: If ``True``, the device should connect to the cloud immediately.
        :param bool overwrite_assets: If ``True``, asset mismatch between the Platform and device definition will be resolved by configuring local assets on the Platform. If ``False``, AssetMismatchException will be raised.

        :raises AssetMismatchException: if asset mismatch is found between the existing asset on the Platform and an asset definition, and overwrite_assets is ``False``

        '''

        self._connected = False

        self.id = id
        self.client = client

        self.overwrite_assets = overwrite_assets
        self.assets = {asset._internal_id: copy.copy(asset) for asset in self._assets}

        def make_get_asset(asset):
            def getter(self):
                return self.client
            return getter

        def make_set_asset(asset):
            def setter(self, value):
                if self._connected:
                    self.client.publish_asset_state(self.id, asset.name, value)
                else:
                    raise RuntimeError('Device not started.')
            return setter

        for asset in self.assets.values():
            asset_property = property(
                make_get_asset(asset), make_set_asset(asset), None,
                asset.description or asset.name or asset._internal_id)
            setattr(type(self), asset._internal_id, asset_property)

        if connect and client:
            self.connect()

    def connect(self, *, client=None, id=None, overwrite_assets=None):
        '''Connects to the device to AllThingsTalk Platform

        :param Client client: The client used to interface with the platform
        :param str id: Device resource id. If supplied, the device will be mapped to the device resource. If None, an attempt will be made to create the device.
        :param bool overwrite_assets: If ``True``, asset mismatch between the Platform and device definition will be resolved by configuring local assets on the Platform. If ``False``, AssetMismatchException will be raised.

        :raises AssetMismatchException: if asset mismatch is found between the existing asset on the Platform and an asset definition, and overwrite_assets is ``False``
        '''

        if id is not None:
            self.id = id
        if client is not None:
            self.client = client
        if overwrite_assets is not None:
            self.overwrite_assets = overwrite_assets

        if not self.id:
            raise NotImplementedError('Device creation not implemented.')

        cloud_assets = {asset['name']: asset for asset in self.client.get_assets(self.id)}
        for asset in self.assets.values():
            name = asset.name
            if name in cloud_assets:
                asset.id = cloud_assets[name]['id']
                asset.thing_id = cloud_assets[name]['deviceId']
            else:
                new_cloud_asset = self.client.create_asset(self.id, asset)
                asset.id = new_cloud_asset['id']
                asset.thing_id = new_cloud_asset['deviceId']

        self.client._attach_device(self)
        self._connected = True

    def _on_message(self, stream, asset_name, message):
        if asset_name in self._handlers[stream]:
            msg = json.loads(message)
            if isinstance(msg, dict):
                msg = {k.lower(): v for k, v in msg.items()}
            else:
                msg = {'value': msg}
            if 'at' in msg and msg['at'] is not None:
                at = parse_date(msg['at'])
            else:
                at = datetime.datetime.utcnow()
            value = msg['value']
            self._handlers[stream][asset_name](self, value, at)



#    _   _ _ _____ _    _              _____     _ _     ___ ___  _  __
#   /_\ | | |_   _| |_ (_)_ _  __ _ __|_   _|_ _| | |__ / __|   \| |/ /
#  / _ \| | | | | | ' \| | ' \/ _` (_-< | |/ _` | | / / \__ \ |) | ' <
# /_/ \_\_|_| |_| |_||_|_|_||_\__, /__/ |_|\__,_|_|_\_\ |___/___/|_|\_\
#                             |___/
#
# Copyright 2017 AllThingsTalk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import profiles


class Asset:

    SENSOR = 'sensor'
    ACTUATOR = 'actuator'
    CONFIG = 'config'
    VIRTUAL = 'virtual'

    _PROFILE_CLASS = None

    def __init__(self, *, kind='sensor', name=None, title=None,
                 description='', handler=None, **kwargs):
        self.id = None
        self.thing_id = None
        self.kind = kind if kind else 'sensor'
        self.name = name
        self._internal_id = name
        self.title = title or name
        self.description = description
        self.handler = handler
        if self.__class__._PROFILE_CLASS:
            self.profile = self.__class__._PROFILE_CLASS(**kwargs)
        elif 'profile' in kwargs and kwargs['profile']:
            self.profile = kwargs['profile']
        if 'type' in self.profile:
            self.type = self.profile['type']


class NumberAsset(Asset):
    _PROFILE_CLASS = profiles.NumberProfile


class IntegerAsset(NumberAsset):
    _PROFILE_CLASS = profiles.IntegerProfile


class StringAsset(Asset):
    _PROFILE_CLASS = profiles.StringProfile


class BooleanAsset(Asset):
    _PROFILE_CLASS = profiles.BooleanProfile


class GeoAsset(Asset):
    _PROFILE_CLASS = profiles.GeoProfile



#    _   _ _ _____ _    _              _____     _ _     ___ ___  _  __
#   /_\ | | |_   _| |_ (_)_ _  __ _ __|_   _|_ _| | |__ / __|   \| |/ /
#  / _ \| | | | | | ' \| | ' \/ _` (_-< | |/ _` | | / / \__ \ |) | ' <
# /_/ \_\_|_| |_| |_||_|_|_||_\__, /__/ |_|\__,_|_|_\_\ |___/___/|_|\_\
#                             |___/
#
# Copyright 2017 AllThingsTalk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import paho.mqtt.client as paho_mqtt
import requests

logger = logging.getLogger('allthingstalk')


class BaseClient:
    '''BaseClient is a base class used for implementing AllThingsTalk Platform
    clients, which are used for interfacing the SDK code with the Platform. It
    doesn't implement any of the client methods.'''

    def _attach_device(self, device):
        pass

    def get_assets(self, device_id):
        raise NotImplementedError('get_assets not implemented')

    def create_asset(self, device_id, asset):
        raise NotImplementedError('create_asset not implemented')

    def publish_asset_state(self, device_id, asset_name, value):
        raise NotImplementedError('publish_asset_state not implemented')


class Client(BaseClient):

    def __init__(self, token, *, api='api.allthingstalk.io',
                 http=None, mqtt=None):
        self.token = token

        def prefix_http(url):
            return url if url.startswith('http') else 'https://%s' % url

        if api.startswith('http://') or api.startswith('https://'):
            api = api.split('//', 1)[1]

        # HTTP Client
        if http:
            self.http = prefix_http(http)
        elif api:
            self.http = prefix_http(api)
        else:
            raise ValueError('Either api or http must be set.')

        # MQTT Client
        if mqtt:
            if ':' in mqtt:
                host, port = mqtt.split(':')[:2]
            else:
                host, port = mqtt, '1883'
            self.mqtt = self._make_mqtt_client(host, port, token)
        elif api:
            host, port = host, port = api, '1883'
            self.mqtt = self._make_mqtt_client(host, port, token)
        else:
            self.mqtt = None

        self._devices = {}

    def _make_mqtt_client(self, host, port, token):
        def on_mqtt_connect(client, userdata, rc):
            logger.debug('MQTT client connected to %s:%s' % (host, port))

        def on_mqtt_disconnect(client, userdata, rc):
            logger.debug('MQTT client disconnected with status %s' % rc)

        def on_mqtt_subscribe(client, userdata, mid, granted_qos):
            pass

        def on_mqtt_message(client, userdata, message):
            logger.debug('MQTT client received a message on topic "%s": %s'
                         % (message.topic, message.payload))
            parts = message.topic.split('/')
            _, device_id, _, asset_name, stream = parts
            self._devices[device_id]._on_message(stream, asset_name, message.payload)

        client = paho_mqtt.Client()
        client.username_pw_set(token, token)
        client.on_connect = on_mqtt_connect
        client.on_disconnect = on_mqtt_disconnect
        client.on_message = on_mqtt_message
        client.on_subscribe = on_mqtt_subscribe
        client.connect(host, int(port), 60)
        client.loop_start()
        return client

    def _attach_device(self, device):
        if self.mqtt:
            logger.debug('Client %s attaching device %s' % (self, device.id))
            for action in ['feed', 'command']:
                logger.debug('Subscribing to %s\'s %ss' % (device.id, action))
                self.mqtt.subscribe('device/%s/asset/+/%s' % (device.id, action))
        self._devices[device.id] = device

    def get_assets(self, device_id):
        return requests.get('%s/device/%s/assets' % (self.http, device_id),
                            headers={'Authorization': 'Bearer %s' % self.token}).json()

    def create_asset(self, device_id, asset):
        attalk_asset = {
            'Name': asset.name,
            'Title': asset.title,
            'Description': asset.description,
            'Is': asset.kind,
            'Profile': asset.profile
        }
        return requests.post('%s/device/%s/assets' % (self.http, device_id),
                             headers={'Authorization': 'Bearer %s' % self.token},
                             json=attalk_asset).json()

    def publish_asset_state(self, device_id, asset_name, value):
        requests.put('%s/device/%s/asset/%s/state' % (self.http, device_id, asset_name),
                     headers={'Authorization': 'Bearer %s' % self.token},
                     json={'value': value})

    def __del__(self):
        try:
            self.mqtt.loop_stop(force=True)
        except:
            pass



from setuptools import setup

setup(
    name='allthingstalk',
    version='0.1.8',
    description='AllThingsTalk Python SDK',
    url='https://github.com/allthingstalk/python-sdk',
    download_url='https://github.com/allthingstalk/python-sdk/archive/0.1.7.tar.gz',
    author='Danilo Vidovic',
    author_email='dv@allthingstalk.com',
    license='Apache 2.0',
    packages=['allthingstalk'],
    keywords=['allthingstalk', 'iot', 'sdk'],
    install_requires=[
        'paho-mqtt==1.2.3',
        'requests',
        'python-dateutil',
        'six'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only'
    ]
)
